#define STRICT
#define _WIN32_WINNT 0x0400
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <string>
#pragma warning( push )
#pragma warning( disable: 4702 )
#include <list>
#pragma warning( pop )
using namespace std;
#include <shellapi.h>
#include <shlobj.h>
#include <objbase.h>
#ifdef UNICODE
#define stlen wcslen
#define stcpy wcscpy
#define stcat wcscat
#define strchr wcsrchr
#define sticmp wcsicmp
#define stcmp wcscmp
#define ststr wcsstr
typedef wstring tstring;
#else
#define stlen strlen
#define stcpy strcpy
#define stcat strcat
#define strchr strrchr
#define sticmp stricmp
#define stcmp strcmp
#define ststr strstr
typedef string tstring;
#endif


// XP Explorer file-opening algorithm:
// (1) HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\FileExts\.ext
//      Application = override.exe
//     This counts as an "override application". If it is present, then the
//     Tools|FileTypes dialog will say "You have customized this filetype".
// (2) HKEY_CLASSES_ROOT\Applications\dummy.exe\shell\open\command = cmdline
//     HKEY_CURRENT_USER\Software\Classes\Applications\dummy.exe\shell\open\command = cmdline
//     These should be the same. They say how to open it with the override.
// (3) HKEY_CLASSES_ROOT\.ext = extfile
//     HKEY_CLASSES_ROOT\extfile\shell\open\command = cmd
//     These say how to open it if no override is present.
//
// To add an intercept, we retrieve the old cmdline (be it override or normal)
// and write ourselves as the override.
// To remove an intercept, we either restore the override to what it was before
// or remove it totally.
//
// Our override applications will be stored under
//    ...\Applications\crlf-interceptor-ext.exe\shell\open\command =
// and this command will be "crlf.exe" "%1" == oldoverride -- oldcmdline
//
// When the user clicks their "restore" button, they will end up removing
// our override as well as their own overrides.

HINSTANCE hInstance=0;
HWND hwnd=0;
tstring fn, name;
HICON hicon=0,hiconsm=0;

class TAutoClose {public: ~TAutoClose() {
  if (hwnd!=0) DestroyWindow(hwnd); hwnd=0;
  if (hicon!=0) DestroyIcon(hicon); hicon=0;
  if (hiconsm!=0) DestroyIcon(hiconsm); hiconsm=0; } } AutoCloser;



int NeedHwnd();

int Error(const TCHAR *fmt, const TCHAR *dat,bool usehwnd=false)
{ const TCHAR *err;
  FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_IGNORE_INSERTS,
    NULL,GetLastError(),MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT),(LPTSTR)&err,0,NULL);
  TCHAR *msg = new TCHAR[stlen(err)+stlen(dat)+stlen(fmt)];
  wsprintf(msg,fmt,err,dat);
  if (usehwnd) NeedHwnd();
  int res = MessageBox(usehwnd?hwnd:NULL,msg,TEXT("Error"),MB_OK|MB_ICONERROR);
  delete[] msg;
  LocalFree((HLOCAL)err);
  return res;
}

int NeedHwnd()
{ if (hwnd!=0) return 1;
  //
  if (fn==TEXT(""))
  { hicon=LoadIcon(hInstance,MAKEINTRESOURCE(1));
    hiconsm=(HICON)LoadImage(hInstance,MAKEINTRESOURCE(1),IMAGE_ICON,GetSystemMetrics(SM_CXSMICON),GetSystemMetrics(SM_CYSMICON),LR_DEFAULTCOLOR);
  }
  else
  { SHFILEINFO shfi; ZeroMemory(&shfi,sizeof(shfi));
    SHGetFileInfo(fn.c_str(),FILE_ATTRIBUTE_NORMAL,&shfi,sizeof(shfi),SHGFI_ICON|SHGFI_USEFILEATTRIBUTES);
    hicon=shfi.hIcon;
    SHGetFileInfo(fn.c_str(),FILE_ATTRIBUTE_NORMAL,&shfi,sizeof(shfi),SHGFI_ICON|SHGFI_SMALLICON|SHGFI_USEFILEATTRIBUTES);
    hiconsm=shfi.hIcon;
  }
  //
  WNDCLASSEX wc; ZeroMemory(&wc,sizeof(wc));
  wc.cbSize=sizeof(wc);
  wc.lpfnWndProc=DefWindowProc;
  wc.hInstance=hInstance;
  wc.hIcon=hicon;
  wc.hIconSm=hiconsm;
  wc.lpszClassName=TEXT("ConvertClass");
  ATOM ares = RegisterClassEx(&wc);
  if (ares==0) return Error(TEXT("Internal error registering window class:\r\n%s%s"),TEXT(""),false);
  tstring tit=TEXT("Convert");
  if (name!=TEXT("")) tit+=TEXT(" - ")+name;
  hwnd=CreateWindowEx(WS_EX_APPWINDOW,TEXT("ConvertClass"),tit.c_str(),WS_VISIBLE|WS_POPUP,0,0,0,0,NULL,0,hInstance,0);
  if (hwnd==0) return Error(TEXT("Internal error creating window:\r\n%s%s"),TEXT(""),false);
  return 1;
}


tstring ExtractFileName(const tstring fn)
{ const TCHAR *lastslash=strchr(fn.c_str(),'\\'); if (lastslash!=0) lastslash++; else
  { lastslash=strchr(fn.c_str(),'/'); if (lastslash!=0) lastslash++; else
    { lastslash=strchr(fn.c_str(),':'); if (lastslash!=0) lastslash++; else lastslash=fn.c_str();
    }
  }
  return tstring(lastslash);
}

tstring ChangeFileExt(const tstring fn,const tstring newext)
{ const TCHAR *lastdot=strchr(fn.c_str(),'.');
  for (const TCHAR *c=lastdot; *c!=0; c++)
  { if (*c=='/' || *c=='\\' || *c==':') return fn+newext;
  }
  const TCHAR *cfn=fn.c_str();
  return tstring(cfn,lastdot-cfn)+newext;
}
  

unsigned int Convert(char *buf, unsigned int size)
{ // PLAN: we will scan forwards counting number of lf, so as to deduce the total
  // file size. Then we scan backwards copying it as necessary.
  int numlf=0;
  for (unsigned int i=0; i<size; i++)
  { if (buf[i]=='\r' && i!=size-1 && buf[i+1]=='\n') i++;
    else if (buf[i]=='\n') numlf++;
  }
  if (numlf==0) return size;
  // we found this many LFs. Therefore, this is how many extra bytes we must add to the file.
  for (unsigned int dst=size+numlf, src=size; src>0; src--,dst--)
  { if (buf[src-1]=='\n' && src>1 && buf[src-2]!='\r') {buf[dst-1]='\n'; buf[dst-2]='\r'; dst--;}
    else buf[dst-1]=buf[src-1];
  }
  if (dst!=0 || src!=0) MessageBox(hwnd,TEXT("There was an internal error converting the file.\r\nI hope nothing was lost! Sorry."),TEXT("Conversion error"),MB_OK);
  return size+numlf;
}




enum TManageTask {mVerify, mAdd, mRemove};

bool Manage(const TCHAR *ext,TManageTask task, const TCHAR *realexe=NULL)
{ //
  // We are going to manage the extension .ext
  //
  // fileclass = (lookup CLASSES_ROOT\.ext)
  //           = extfile
  HKEY key=NULL; TCHAR dat[10000]; DWORD size,type,disp;
  LONG res=RegOpenKeyEx(HKEY_CLASSES_ROOT,ext,0,KEY_READ,&key);
  if (res!=ERROR_SUCCESS) return false;
  size=sizeof(dat); *dat=0; res=RegQueryValueEx(key,NULL,NULL,&type,(LPBYTE)&dat,&size);
  RegCloseKey(key);
  if (res!=ERROR_SUCCESS) return false;
  tstring fileclass(dat);
  //
  //
  // old_defcmd = (lookup CLASSES_ROOT\.ext\shell\open\command)
  //            = "c:\program files\msdev.exe" "%1"
  tstring cmdkey=fileclass+TEXT("\\shell\\open\\command");
  tstring old_defcmd;
  res=RegOpenKeyEx(HKEY_CLASSES_ROOT,cmdkey.c_str(),0,KEY_READ,&key);
  if (res==ERROR_SUCCESS)
  { size=sizeof(dat); *dat=0; res=RegQueryValueEx(key,NULL,NULL,&type,(LPBYTE)&dat,&size);
    RegCloseKey(key);
    old_defcmd=dat;
  }
  else
  { // maybe there wasn't an "open" verb, but there was some other verb at least...
    cmdkey=fileclass+TEXT("\\shell");
    res=RegOpenKeyEx(HKEY_CLASSES_ROOT,cmdkey.c_str(),0,KEY_READ,&key);
    if (res==ERROR_SUCCESS)
    { size=sizeof(dat); *dat=0; FILETIME ft;
      res=RegEnumKeyEx(key,0,dat,&size,NULL,NULL,NULL,&ft);
      if (res==ERROR_SUCCESS)
      { HKEY key2; tstring cmdkey=fileclass+TEXT("\\shell\\")+dat+TEXT("\\command");
        res=RegOpenKeyEx(HKEY_CLASSES_ROOT,cmdkey.c_str(),0,KEY_READ,&key2);
        if (res==ERROR_SUCCESS)
        { size=sizeof(dat); *dat=0; res=RegQueryValueEx(key2,NULL,NULL,&type,(LPBYTE)&dat,&size);
          old_defcmd=dat;
          RegCloseKey(key2);
        }
      }
    }
  }
  if (old_defcmd.length()==0) old_defcmd=TEXT("\"notepad.exe\" \"%1\"");
  //
  //
  // old_override = (lookup CURRENT_USER\...\FileExts\.ext)
  //              = "notepad.exe"
  //              = "crlf-interceptor-ext.exe"
  //              = empty if there were no overrides before
  cmdkey=TEXT("Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\FileExts\\")+tstring(ext);
  tstring old_override; 
  res=RegOpenKeyEx(HKEY_CURRENT_USER,cmdkey.c_str(),0,KEY_READ,&key);
  if (res==ERROR_SUCCESS)
  { size=sizeof(dat); *dat=0; res=RegQueryValueEx(key,TEXT("Application"),NULL,&type,(LPBYTE)&dat,&size);
    RegCloseKey(key);
    old_override = tstring(dat);
  }
  //
  //
  // old_overridecmd = (lookup ..\Applications\old_override\shell\open\command)
  //                 = "c:\progra~1\crlf-interceptor.exe" "%1" == notepad.exe -- notepad.exe "%1"
  //                 = notepad.exe "%1"
  //                 = "c:\progra~1\crlf-interceptor.exe" "%1" -- msdev.exe "%1"
  tstring old_overridecmd;
  if (old_override.length()>0)
  { cmdkey=TEXT("Applications\\")+old_override+TEXT("\\shell\\open\\command");
    res=RegOpenKeyEx(HKEY_CLASSES_ROOT,cmdkey.c_str(),0,KEY_READ,&key);
    if (res==ERROR_SUCCESS)
    { size=sizeof(dat); *dat=0; res=RegQueryValueEx(key,NULL,NULL,&type,(LPBYTE)&dat,&size);
      RegCloseKey(key);
      old_overridecmd = tstring(dat);
    }
  }
  //
  //
  // old_cmd = (either old_overridecmd or old_defcmd)
  //         = "c:\progra~1\crlf-interceptor.exe" "%1" == notepad.exe -- notepad.exe "%1"
  //         = "c:\progra~1\crlf-interceptor.exe" "%1" -- notepad.exe "%1"
  //         = notepad.exe "%1"
  tstring old_cmd;
  if (old_overridecmd.length()>0) old_cmd = old_overridecmd;
  else old_cmd = old_defcmd;
  //
  //
  // old_exepath, old_exename = (the primary argument in old_cmd)
  //                          = c:\progra~1\crlf-interceptor.exe  /  crlf-interceptor.exe
  //                          = notepad.exe / notepad.exe
  const TCHAR *start=old_cmd.c_str(), *end;
  if (*start=='\"')
  { start++; if (*start==0) return false; end=start+1;
    while (*end!='\"' && *end!=0) end++;
  }
  else
  { end=start; while (*end!=' ' && *end!=0) end++;
  }
  tstring old_exepath(start,end-start), old_exename=ExtractFileName(old_exepath);
  //
  //
  // this_exepath, this_exename, thisexebase = (the current executable)
  //                            = c:\progra~1\crlf-interceptor.exe  /  crlf-interceptor.exe
  tstring this_exepath;
  if (realexe==NULL)
  { TCHAR buf[MAX_PATH]; GetModuleFileName(hInstance,buf,MAX_PATH);
    this_exepath=tstring(buf);
  }
  else this_exepath=tstring(realexe);
  tstring this_exename = ExtractFileName(this_exepath);
  tstring this_exebase = ChangeFileExt(this_exename,TEXT(""));
  //
  //
  // isthisexe = (whether we are being used to execute the file)
  bool thisexe = (sticmp(this_exename.c_str(),old_exename.c_str())==0);
  if (task==mVerify) return thisexe;
  //
  //
  // db = (our private list of the things we think we handle)
  //      (including the addition/removal of .ext)
  list<tstring> db; bool mustaddthis=true;
  res=RegOpenKeyEx(HKEY_CLASSES_ROOT,TEXT("Software\\Lu\\Crlf-interceptor"),0,KEY_READ,&key);
  if (res==ERROR_SUCCESS)
  { size=sizeof(dat); *dat=0; res=RegQueryValueEx(key,TEXT("managed"),NULL,&type,(LPBYTE)&dat,&size);
    RegCloseKey(key);
    if (res==ERROR_SUCCESS)
    { const TCHAR *c=dat; while (*c!=0)
      { const TCHAR *start=c; while (*c!=0 && *c!=';') c++;
        if (start==c) break;
        tstring dbext(start,c-start);
        bool canhave=true;
        bool isthis = sticmp(ext,dbext.c_str())==0;
        if (isthis) mustaddthis=false;
        if (task==mRemove && isthis) canhave = false;
        if (canhave) db.push_back(dbext);
        if (*c!=0) c++;
      }
    }
  }
  if (task==mAdd && mustaddthis) db.push_back(ext);
  //
  // 
  // Write our modified db.
  // We write the database before writing our override because it would be yucky
  // if the override had been written but our database write had failed. (i.e.
  // we wouldn't be able to restore it ourselves.)
  tstring dbs;
  for (list<tstring>::const_iterator i=db.begin(); i!=db.end(); i++)
  { if (dbs!=TEXT("")) dbs=dbs+TEXT(";")+*i; else dbs=dbs+*i;
  }
  res=RegCreateKeyEx(HKEY_CLASSES_ROOT,TEXT("Software\\Lu\\Crlf-interceptor"),0,NULL,0,KEY_ALL_ACCESS,NULL,&key,&disp);
  if (res!=ERROR_SUCCESS) return false;
  size=sizeof(TCHAR)*((DWORD)dbs.length()+1); res=RegSetValueEx(key,TEXT("managed"),0,REG_SZ,(CONST BYTE*)dbs.c_str(),size);
  RegCloseKey(key);
  if (res!=ERROR_SUCCESS) return false;
  //
  //
  // newcmd = "\progra~1\crlf.exe" "%1" == oldoverride -- oldcmdline
  tstring newcmd = TEXT("\"")+this_exepath+TEXT("\" \"%1\"");
  if (old_override.length()>0) newcmd+=TEXT(" == ")+old_override;
  newcmd += TEXT(" -- ")+old_cmd;
  //
  //
  // If adding, then we have to create the ...Applications\crlf-ext.exe\shell\open\command
  // If removing, we have to delete them.
  if (task==mAdd && !thisexe)
  { cmdkey = TEXT("Applications\\")+this_exebase+TEXT("-")+tstring(ext+1)+TEXT(".exe\\shell\\open\\command");
    res=RegCreateKeyEx(HKEY_CLASSES_ROOT,cmdkey.c_str(),0,NULL,0,KEY_ALL_ACCESS,NULL,&key,&disp);
    if (res!=ERROR_SUCCESS) return false;
    size=sizeof(TCHAR)*((DWORD)newcmd.length()+1);
    res=RegSetValueEx(key,NULL,0,REG_EXPAND_SZ,(CONST BYTE*)newcmd.c_str(),size);
    RegCloseKey(key);
    if (res!=ERROR_SUCCESS) return false;
    cmdkey = TEXT("Software\\Classes\\")+cmdkey;
    res=RegCreateKeyEx(HKEY_CLASSES_ROOT,cmdkey.c_str(),0,NULL,0,KEY_ALL_ACCESS,NULL,&key,&disp);
    if (res!=ERROR_SUCCESS) return false;
    size=sizeof(TCHAR)*((DWORD)newcmd.length()+1);
    res=RegSetValueEx(key,NULL,0,REG_EXPAND_SZ,(CONST BYTE*)newcmd.c_str(),size);
    RegCloseKey(key);
    if (res!=ERROR_SUCCESS) return false;
  }
  if (task==mRemove)
  { cmdkey = TEXT("Applications\\")+this_exebase+TEXT("-")+tstring(ext+1)+TEXT(".exe");
    RegDeleteKey(HKEY_CLASSES_ROOT,(cmdkey+TEXT("\\shell\\open\\command")).c_str());
    RegDeleteKey(HKEY_CLASSES_ROOT,(cmdkey+TEXT("\\shell\\open")).c_str());
    RegDeleteKey(HKEY_CLASSES_ROOT,(cmdkey+TEXT("\\shell")).c_str());
    RegDeleteKey(HKEY_CLASSES_ROOT,(cmdkey).c_str());
    cmdkey = TEXT("Software\\Classes\\")+cmdkey;
    RegDeleteKey(HKEY_CURRENT_USER,(cmdkey+TEXT("\\shell\\open\\command")).c_str());
    RegDeleteKey(HKEY_CURRENT_USER,(cmdkey+TEXT("\\shell\\open")).c_str());
    RegDeleteKey(HKEY_CURRENT_USER,(cmdkey+TEXT("\\shell")).c_str());
    RegDeleteKey(HKEY_CURRENT_USER,(cmdkey).c_str());
  }
  //
  //
  // If adding, we must present ourselves as the override
  // If removing, we must figure out the old override if there was one,
  // and either set it as the override or remove the override if there wasn't.
  tstring new_override=TEXT("*");
  if (task==mAdd && !thisexe)
  { new_override = this_exebase+TEXT("-")+tstring(ext+1)+TEXT(".exe");
  }
  if (task==mRemove && thisexe)
  { // old_cmd = "c:\progra~1\crlf-interceptor.exe" "%1" == notepad.exe -- notepad.exe "%1"
    // or      = "c:\progra~1\crlf-interceptor.exe" "%1" -- notepad.exe "%1"
    // in the first case, we set new_override = notepad.exe
    // in the second case, we set new_override = (empty)
    const TCHAR *oldc = old_cmd.c_str() + old_exepath.length();
    // that way, oldc will be at or near the end of the old exe name.
    // which avoids an (unlikely) bug if the old exe name had included -- or ==
    const TCHAR *slashes = ststr(oldc,TEXT(" -- "));
    const TCHAR *equals  = ststr(oldc,TEXT(" == "));
    if (equals!=0 && equals<slashes) new_override=tstring(equals+4,slashes-equals-4);
    else new_override=TEXT("");
  }
  if (new_override!=TEXT("*"))
  { cmdkey=TEXT("Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\FileExts\\")+tstring(ext);
    res=RegCreateKeyEx(HKEY_CURRENT_USER,cmdkey.c_str(),0,NULL,0,KEY_ALL_ACCESS,NULL,&key,&disp);
    if (res!=ERROR_SUCCESS) return false;
    if (new_override.length()==0) res = RegDeleteValue(key,TEXT("Application"));
    else
    { size=sizeof(TCHAR)*((DWORD)new_override.length()+1);
      res=RegSetValueEx(key,TEXT("Application"),0,REG_SZ,(CONST BYTE*)new_override.c_str(),size);
      RegCloseKey(key);
    }
    if (res!=ERROR_SUCCESS) return false;
  }
  //
  //
  return true;
}

list<tstring> GetManageds()
{ list<tstring> m;
  HKEY key=NULL; TCHAR dat[10000]; DWORD size; DWORD type;
  LONG res=RegOpenKeyEx(HKEY_CLASSES_ROOT,TEXT("Software\\Lu\\Crlf-interceptor"),0,KEY_READ,&key);
  if (res!=ERROR_SUCCESS) return m;
  size=sizeof(dat); *dat=0; res=RegQueryValueEx(key,TEXT("managed"),NULL,&type,(LPBYTE)&dat,&size);
  RegCloseKey(key);
  const TCHAR *c=dat; while (*c!=0)
  { const TCHAR *start=c; while (*c!=0 && *c!=';') c++;
    if (start==c) return m;
    tstring ext(start,c-start);
    bool isreal = Manage(ext.c_str(),mVerify);
    if (isreal) m.push_back(ext);
    if (*c!=0) c++;
  }
  // Must also validate the managed item to see if it really is
  return m;
}
   

BOOL CALLBACK NameDlgProc(HWND hdlg,UINT msg,WPARAM wParam,LPARAM lParam)
{ switch (msg)
  { case WM_INITDIALOG:
    { SetWindowLongPtr(hdlg,DWLP_USER,(LONG)lParam);
      TCHAR *c = (TCHAR*)lParam;
      SetDlgItemText(hdlg,101,c);
      return TRUE;
    }
    case WM_COMMAND:
    { int id=LOWORD(wParam), code=HIWORD(wParam);
      if (id==IDOK && code==BN_CLICKED)
      { TCHAR *c = (TCHAR*)(LONG_PTR)GetWindowLongPtr(hdlg,DWLP_USER);
        GetDlgItemText(hdlg,101,c,MAX_PATH);
        EndDialog(hdlg,IDOK);
      }
      if (id==IDCANCEL && code==BN_CLICKED) EndDialog(hdlg,IDCANCEL);
      return TRUE;
    }
  }
  return FALSE;
}


BOOL CALLBACK ConfigDlgProc(HWND hdlg,UINT msg,WPARAM wParam,LPARAM)
{ switch (msg)
  { case WM_INITDIALOG:
    { RECT rc; GetWindowRect(hdlg,&rc); int width=rc.right-rc.left, height=rc.bottom-rc.top;
      int x=(GetSystemMetrics(SM_CXSCREEN)-width)/2, y=(GetSystemMetrics(SM_CYSCREEN)-height)/2;
      MoveWindow(hdlg,x,y,width,height,FALSE);
      list<tstring> m = GetManageds();
      SendDlgItemMessage(hdlg,102,LB_RESETCONTENT,0,0);
      for (list<tstring>::const_iterator i=m.begin(); i!=m.end(); i++)
      { SendDlgItemMessage(hdlg,102,LB_ADDSTRING,0,(LPARAM)i->c_str());
      }
      EnableWindow(GetDlgItem(hdlg,104),FALSE);
    }
    case WM_COMMAND:
    { int id=LOWORD(wParam), code=HIWORD(wParam);
      if (id==102 && code==LBN_SELCHANGE)
      { LRESULT i = SendDlgItemMessage(hdlg,102,LB_GETCURSEL,0,0);
        EnableWindow(GetDlgItem(hdlg,104), i!=LB_ERR);
      }
      if (id==103 && code==BN_CLICKED)
      { TCHAR c[MAX_PATH]; stcpy(c,TEXT(".txt"));
        int res = (int)DialogBoxParam(hInstance,MAKEINTRESOURCE(3),hdlg,NameDlgProc,(LPARAM)c);
        if (res==IDOK && *c!=0)
        { tstring s(c); if (s[0]!='.') s=TEXT(".")+s;
          LRESULT i = SendDlgItemMessage(hdlg,102,LB_FINDSTRINGEXACT,(WPARAM)-1,(LPARAM)s.c_str());
          if (i==LB_ERR) SendDlgItemMessage(hdlg,102,LB_ADDSTRING,0,(LPARAM)s.c_str());
        }
      }
      if (id==104 && code==BN_CLICKED)
      { LRESULT i = SendDlgItemMessage(hdlg,102,LB_GETCURSEL,0,0);
        if (i!=LB_ERR)
        { SendDlgItemMessage(hdlg,102,LB_DELETESTRING,i,0);
          SendDlgItemMessage(hdlg,102,LB_SETCURSEL,(WPARAM)-1,0);
          EnableWindow(GetDlgItem(hdlg,104),FALSE);
        }
      }
      if (id==IDOK && code==BN_CLICKED)
      { list<tstring> oldm = GetManageds();
        list<tstring> newm;
        int num = (int)SendDlgItemMessage(hdlg,102,LB_GETCOUNT,0,0);
        for (int i=0; i<num; i++)
        { TCHAR buf[MAX_PATH]; SendDlgItemMessage(hdlg,102,LB_GETTEXT,i,(LPARAM)buf);
          newm.push_back(tstring(buf));
        }
        oldm.sort(); newm.sort();
        list<tstring>::const_iterator oldi=oldm.begin(), newi=newm.begin();
        tstring errors; int numerrors=0;
        for (;;)
        { const TCHAR *toadd=0, *toremove=0;
          if (oldi==oldm.end() && newi==newm.end()) break; // end of both lists
          else if (oldi==oldm.end()) {toadd=newi->c_str(); newi++;}
          else if (newi==newm.end()) {toremove=oldi->c_str(); oldi++;}
          else if (stcmp(oldi->c_str(),newi->c_str())>0) {toadd=newi->c_str(); newi++;}
          else if (stcmp(oldi->c_str(),newi->c_str())<0) {toremove=oldi->c_str(); oldi++;}
          else if (stcmp(oldi->c_str(),newi->c_str())==0) {newi++; oldi++;}
          else MessageBox(hdlg,TEXT("Internal error! Lost track changes made in this dialog"),TEXT("Error"),MB_OK);
          bool ok=true;
          if (toremove!=0) ok=Manage(toremove,mRemove);
          if (toadd!=0) ok=Manage(toadd,mAdd);
          if (!ok) {numerrors++; if (errors!=TEXT("")) errors+=TEXT(";"); errors+=tstring(toremove)+tstring(toadd);}
        }
        if (errors!=TEXT(""))
        { tstring msg=TEXT("Internal error! There were problems adjusting ");
          if (numerrors==1) msg+=errors;
          else msg+=TEXT("these extensions:\r\n")+errors;
          MessageBox(hdlg,msg.c_str(),TEXT("Error"),MB_OK|MB_ICONEXCLAMATION);
        }
        EndDialog(hdlg,IDOK);
      }
      if (id==IDCANCEL && code==BN_CLICKED) EndDialog(hdlg,IDCANCEL);
      return TRUE;
    }
  }
  return FALSE;
}



TCHAR initialdir[MAX_PATH];
//
INT CALLBACK BrowseCallbackProc(HWND hwnd,UINT uMsg,LPARAM lp,LPARAM)
{ switch(uMsg)
  { case BFFM_INITIALIZED:
    { if (initialdir[0]!=0) SendMessage(hwnd,BFFM_SETSELECTION,TRUE,(LPARAM)initialdir);
    } break;
    case BFFM_SELCHANGED: // Set the status window to the currently selected path.
    { TCHAR c[MAX_PATH];
      if (SHGetPathFromIDList((LPITEMIDLIST) lp ,c)) SendMessage(hwnd,BFFM_SETSTATUSTEXT,0,(LPARAM)c);
    } break;
  }
  return 0;
}



BOOL CALLBACK InstallDlgProc(HWND hdlg,UINT msg,WPARAM wParam,LPARAM)
{ switch (msg)
  { case WM_INITDIALOG:
    { CoInitializeEx(NULL,COINIT_APARTMENTTHREADED);
      RECT rc; GetWindowRect(hdlg,&rc); int width=rc.right-rc.left, height=rc.bottom-rc.top;
      int x=(GetSystemMetrics(SM_CXSCREEN)-width)/2, y=(GetSystemMetrics(SM_CYSCREEN)-height)/2;
      MoveWindow(hdlg,x,y,width,height,FALSE);
      IMalloc *shmalloc; SHGetMalloc(&shmalloc);
      ITEMIDLIST *pidl; HRESULT res = SHGetSpecialFolderLocation(NULL,CSIDL_PROGRAM_FILES,&pidl);
      TCHAR dpath[MAX_PATH]; if (res==NOERROR) SHGetPathFromIDList(pidl,dpath);
      else stcpy(dpath,TEXT("c:\\Program Files"));
      shmalloc->Free(pidl);
      shmalloc->Release();
      SetDlgItemText(hdlg,105,dpath);
      CheckDlgButton(hdlg,107,BST_CHECKED);
      return TRUE;
    }
    case WM_DESTROY:
    { CoUninitialize();
      return FALSE;
    }
    case WM_COMMAND:
    { int id=LOWORD(wParam), code=HIWORD(wParam);
      if (id==106 && code==BN_CLICKED)
      { TCHAR dname[MAX_PATH];
        GetDlgItemText(hdlg,105,initialdir,MAX_PATH);
        IMalloc *imalloc; SHGetMalloc(&imalloc);
        BROWSEINFO bi; ZeroMemory(&bi,sizeof(bi));
        bi.hwndOwner=hdlg;
        bi.pszDisplayName=dname;
        bi.lpszTitle = TEXT("Choose a folder");
        bi.ulFlags = BIF_USENEWUI|BIF_NONEWFOLDERBUTTON|BIF_RETURNONLYFSDIRS;
        bi.lpfn = BrowseCallbackProc;
        ITEMIDLIST *pidl = SHBrowseForFolder(&bi);
        if (pidl!=NULL)
        { TCHAR dir[MAX_PATH]; SHGetPathFromIDList(pidl,dir);
          SetDlgItemText(hdlg,105,dir);
        }
        imalloc->Free(pidl);
        imalloc->Release();
      }
      if (id==IDOK && code==BN_CLICKED)
      { TCHAR c[MAX_PATH]; GetDlgItemText(hdlg,105,c,MAX_PATH);
        tstring ifn(c); ifn+=TEXT("\\crlf-interceptor.exe");
        // copy the file into the program-files directory
        TCHAR src[MAX_PATH]; GetModuleFileName(hInstance,src,MAX_PATH);
        CopyFile(src,ifn.c_str(),FALSE);
        // There's just one subtlety here. If we the user uninstalled it,
        // and then immediately reinstalled it, then upon reboot it would still be
        // deleted. Oh well, nothing we can do about that.
        // make an "uninstall" key
        LONG lres; HKEY skey; DWORD disp; 
        lres=RegCreateKeyEx(HKEY_LOCAL_MACHINE,TEXT("SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\Crlf-interceptor"),0,NULL,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&skey,&disp);
        if (lres==ERROR_SUCCESS)
        { tstring val=TEXT("Crlf Interceptor Unix->Windows"); RegSetValueEx(skey,TEXT("DisplayName"),0,REG_SZ,(const BYTE*)val.c_str(),sizeof(TCHAR)*((DWORD)val.length()+1));
          val=TEXT("\"")+ifn+TEXT("\" /uninstall"); RegSetValueEx(skey,TEXT("UninstallString"),0,REG_SZ,(const BYTE*)val.c_str(),sizeof(TCHAR)*((DWORD)val.length()+1));
          RegCloseKey(skey);
        }
        // register the filetype
        int chk = IsDlgButtonChecked(hdlg,107);
        if (chk==BST_CHECKED) Manage(TEXT(".txt"),mAdd,ifn.c_str());
        // add to start menu
        IMalloc *shmalloc; SHGetMalloc(&shmalloc);
        ITEMIDLIST *pidl; HRESULT hres = SHGetSpecialFolderLocation(NULL,CSIDL_PROGRAMS,&pidl);
        TCHAR dpath[MAX_PATH]; SHGetPathFromIDList(pidl,dpath);
        tstring shpath(dpath); shpath+=TEXT("\\Crlf Interceptor.lnk");
        if (hres==NOERROR)
        { IShellLink *shortcut;
          hres = CoCreateInstance(CLSID_ShellLink,NULL,CLSCTX_INPROC_SERVER,IID_IShellLink,(void **)&shortcut);      
          if (SUCCEEDED(hres))
          { IPersistFile *ppf;
            hres = shortcut->QueryInterface(IID_IPersistFile,(void**)&ppf);
            if (SUCCEEDED(hres))
            { shortcut->SetPath(ifn.c_str());
              shortcut->SetDescription(TEXT("Crlf Interceptor"));
              #ifdef UNICODE
              ppf->Save(shpath.c_str(), TRUE);
              #else
              wchar_t ws[MAX_PATH];
              MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,shpath.c_str(),-1,ws,MAX_PATH);
              ppf->Save(ws,TRUE);
              #endif
            }
            ppf->Release();
          }
          shortcut->Release();
        }
        shmalloc->Free(pidl);
        shmalloc->Release();
        // finished!
        if (chk==BST_CHECKED) EndDialog(hdlg,IDOK);
        else EndDialog(hdlg,IDNO);
      }
      if (id==IDCANCEL && code==BN_CLICKED) EndDialog(hdlg,IDCANCEL);
      return TRUE;
    }
  }
  return FALSE;
}

int Install()
{ NeedHwnd(); int res = (int)DialogBox(hInstance,MAKEINTRESOURCE(1),hwnd,InstallDlgProc);
  if (res==IDCANCEL) return res;
  tstring msg;
  if (res==IDOK) msg=TEXT("Install completed.\r\nClick on a unixy text file to watch it be converted!");
  else msg=TEXT("Install completed.\r\nConfigure file-types from the \"Crlf Interceptor\" icon in the Start menu.");
  MessageBox(hwnd,msg.c_str(),TEXT("Crlf Interceptor"),MB_OK);
  return res;
}




int Uninstall()
{ // First, go through the database removing everything
  list<tstring> db=GetManageds();
  for (list<tstring>::const_iterator i=db.begin(); i!=db.end(); i++)
  { Manage(i->c_str(),mRemove);
  }
  // Second, remove registry key for the program
  RegDeleteKey(HKEY_CLASSES_ROOT,TEXT("Software\\Lu\\Crlf-interceptor"));
  // Third, remove uninstall registry key
  RegDeleteKey(HKEY_LOCAL_MACHINE,TEXT("SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\Crlf-interceptor"));
  // Fourth, remove from start menu
  IMalloc *shmalloc; SHGetMalloc(&shmalloc);
  ITEMIDLIST *pidl; SHGetSpecialFolderLocation(NULL,CSIDL_PROGRAMS,&pidl);
  TCHAR dpath[MAX_PATH]; SHGetPathFromIDList(pidl,dpath);
  tstring shpath(dpath); shpath+=TEXT("\\Crlf Interceptor.lnk");
  DWORD attr=GetFileAttributes(shpath.c_str());
  if (attr!=0xFFFFFFFF) DeleteFile(shpath.c_str());
  shmalloc->Free(pidl);
  shmalloc->Release();
  // Fifth, tell the executable to disappear upon next reboot of windows
  TCHAR fn[MAX_PATH]; GetModuleFileName(hInstance,fn,MAX_PATH);
  bool mustwininit=true;
  OSVERSIONINFO osv; ZeroMemory(&osv,sizeof(osv)); osv.dwOSVersionInfoSize=sizeof(osv);
  GetVersionEx(&osv);
  bool canwininit = true; //(osv.dwPlatformId==VER_PLATFORM_WIN32_WINDOWS);
  //
  HINSTANCE hlib = LoadLibrary(TEXT("kernel32.dll"));
  typedef BOOL (WINAPI *MOVEFILEEX)(LPCTSTR,LPCTSTR,DWORD);
  MOVEFILEEX pMoveFileEx = (MOVEFILEEX)GetProcAddress(hlib,"MoveFileExA");
  if (pMoveFileEx!=0)
  { BOOL res = pMoveFileEx(fn,NULL,MOVEFILE_DELAY_UNTIL_REBOOT);
    if (res) mustwininit=false;
  }
  FreeLibrary(hlib);
  if (mustwininit && !canwininit)
  { NeedHwnd();
    tstring msg(TEXT("There was a problem uninstall.ing.\r\nTo complete the uninstall manually, you should delete the file\r\n"));
    msg += tstring(fn);
    MessageBox(hwnd,msg.c_str(),TEXT("Crlf Interceptor Uninstall"),MB_OK);
    return 0;
  }
  if (mustwininit)
  { TCHAR windir[MAX_PATH]; GetWindowsDirectory(windir,MAX_PATH);
    TCHAR sfn[MAX_PATH]; GetShortPathName(fn,sfn,MAX_PATH);
    tstring ini=tstring(windir)+TEXT("\\wininit.ini");
    WritePrivateProfileString(TEXT("rename"),TEXT("NUL"),sfn,ini.c_str());
  }
  //
  //
  NeedHwnd();
  MessageBox(hwnd,TEXT("Uninstall completed.\r\nThe last traces of the program will be removed when next you reboot.\r\n(So don't try reinstalling it without first rebooting!)"),TEXT("Crlf Interceptor Uninstall"),MB_OK);
  return 0;
}


int WINAPI WinMain(HINSTANCE hinst,HINSTANCE,LPSTR,int)
{ hInstance=hinst;
  const TCHAR *arg = GetCommandLine(); bool issetup=false;
  if (*arg=='\"')
  { arg++;
    while (*arg!='\"' && *arg!=0)
    { issetup |= ((arg[0]=='s'||arg[0]=='S')&&(arg[1]=='e'||arg[1]=='E')&&(arg[2]=='t'||arg[2]=='T')&&(arg[3]=='u'||arg[3]=='U')&&(arg[4]=='p'||arg[4]=='P'));
      arg++;
    }
    if (*arg=='\"') arg++;
  }
  else
  { while (*arg!=' ')
    { issetup |= ((arg[0]=='s'||arg[0]=='S')&&(arg[1]=='e'||arg[1]=='E')&&(arg[2]=='t'||arg[2]=='T')&&(arg[3]=='u'||arg[3]=='U')&&(arg[4]=='p'||arg[4]=='P'));
      arg++;
    }
  }
  while (*arg==' ') arg++;
  if (issetup) return Install();
  if (*arg==0) {NeedHwnd(); return (int)DialogBox(hInstance,MAKEINTRESOURCE(2),hwnd,ConfigDlgProc);}
  if (sticmp(arg,TEXT("/uninstall"))==0) return Uninstall();
  //
  const TCHAR *rest;
  if (*arg=='\"')
  { const TCHAR *start = arg+1, *c=start;
    while (*c!='\"' && *c!=0) c++;
    fn=tstring(start,c-start);
    if (*c=='\"') rest=c+1; else rest=c;
  }
  else
  { const TCHAR *start=arg, *c=arg;
    while (*c!=' ' && *c!=0) c++;
    fn=tstring(start,c-start);
    rest=c;
  }
  name=ExtractFileName(fn);
  //
  bool skiptest = (GetAsyncKeyState(VK_SHIFT)<0) || (GetAsyncKeyState(VK_LSHIFT)<0) || (GetAsyncKeyState(VK_RSHIFT)<0);
  bool cancel=false;
  if (*arg!=0 && !skiptest)
  { HANDLE hf = CreateFile(fn.c_str(),GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,0,NULL);
    if (hf!=INVALID_HANDLE_VALUE)
    { DWORD sizehigh, size=GetFileSize(hf,&sizehigh);
      if (sizehigh==0 && size>2 && size<0x00FFFFFF)
      { if (size>1000) size=1000; unsigned char tst[1000]; tst[size-1]=0;
        DWORD red; ReadFile(hf,tst,size-1,&red,NULL); CloseHandle(hf);
        int numtext=0, numbin=0, numlf=0, numcrlf=0, numhightext=0;
        for (unsigned int i=0; i<size-2; i++)
        { unsigned char c = tst[i];
          if (c>=32 && c<=127) numtext++;
          else if (c>=128 && c<=255) numhightext++;
          else if (c=='\t' || c=='\f') numtext++;
          else if (c=='\r' && tst[i+1]=='\n') {numcrlf++; i++;}
          else if (c=='\n') numlf++;
          else numbin++;
        }
        // heuristics for whether it's a unix file
        bool isunix = (numlf>1 && numcrlf==0 && numbin==0 && numhightext<numtext/10 && numtext>5);
        if (isunix)
        { tstring msg = TEXT("The file ")+name+TEXT(" seems to be a unix text file.\r\nDo you want to convert it to Windows?");
          NeedHwnd();
          int res = MessageBox(hwnd,msg.c_str(),TEXT("Convert text file?"),MB_YESNOCANCEL);
          bool wasreadonly=false;
          if (res==IDYES)
          { DWORD attr = GetFileAttributes(fn.c_str());
            if (attr&FILE_ATTRIBUTE_READONLY)
            { tstring msg = TEXT("The file ")+name+TEXT(" is read-only.\r\nAre you sure you want to convert it?");
              res = MessageBox(hwnd,msg.c_str(),TEXT("Convert text file?"),MB_YESNOCANCEL);
              if (res==IDYES)
              { attr &= ~FILE_ATTRIBUTE_READONLY;
                SetFileAttributes(fn.c_str(),attr);
                wasreadonly=true;
              }
            }
          }
          if (res==IDCANCEL) cancel=true;
          if (res==IDYES)
          { HANDLE hf = CreateFile(fn.c_str(),GENERIC_READ|GENERIC_WRITE,0,NULL,OPEN_EXISTING,0,NULL);
            if (hf==INVALID_HANDLE_VALUE) Error(TEXT("Unable to open file:\r\n%s\r\nThe file was\r\n%s"),fn.c_str());
            else
            { DWORD size=GetFileSize(hf,NULL), newsize=size;
              HANDLE hmap = CreateFileMapping(hf,NULL,PAGE_READWRITE,0,size*2,NULL);
              if (hmap==NULL) Error(TEXT("Unable to create file mapping:\r\n%s\r\nThe file was\r\n%s"),fn.c_str());
              else
              { char *buf = (char*)MapViewOfFile(hmap,FILE_MAP_WRITE,0,0,size*2);
                if (buf==0) Error(TEXT("Unable to map file:\r\n%s\r\nThe file was\r\n%s"),fn.c_str());
                else
                { newsize = Convert(buf,size);
                  UnmapViewOfFile(buf);
                }
                CloseHandle(hmap);
              }
              SetFilePointer(hf,newsize,NULL,FILE_BEGIN); 
              SetEndOfFile(hf); // truncates it!
              CloseHandle(hf);
              if (wasreadonly)
              { DWORD attr = GetFileAttributes(fn.c_str());
                attr|=FILE_ATTRIBUTE_READONLY;
                SetFileAttributes(fn.c_str(),attr);
              }
            }
          }
        }
      }
    }
  }
  if (!cancel)
  { // arg points to the string... "%1" == defapp -- oldcmd
    // we can simply use oldcmd, since its %1s will already have been substituted!
    tstring newcmd; 
    for (const TCHAR *slashes=rest; *slashes!=0; slashes++)
    { bool isslash = (slashes[0]==' ' && slashes[1]=='-' && slashes[2]=='-' && slashes[3]==' ');
      if (isslash) {newcmd=tstring(slashes+4); break;}
    }
    if (newcmd.length()==0) newcmd = TEXT("\"notepad.exe\" \"")+fn+TEXT("\"");
    STARTUPINFO si; ZeroMemory(&si,sizeof(si)); si.cb=sizeof(si);
    PROCESS_INFORMATION pi; ZeroMemory(&pi,sizeof(pi));
    BOOL res = CreateProcess(NULL,(TCHAR*)newcmd.c_str(),NULL,NULL,FALSE,0,NULL,NULL,&si,&pi);
    if (res) {CloseHandle(pi.hProcess); CloseHandle(pi.hThread);}
    else Error(TEXT("Unable to launch notepad:\r\n%s\r\nThe command line was\r\n%s"),newcmd.c_str());
  }
  return 0;
}