#define STRICT
#define _WIN32_WINNT 0x0400
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <string>
#pragma warning( push )
#pragma warning( disable: 4702 )
#include <list>
#pragma warning( pop )
using namespace std;
#include <shellapi.h>
#include <shlobj.h>
#include <objbase.h>
#ifdef UNICODE
#define stlen wcslen
#define stcpy wcscpy
#define stcat wcscat
#define strchr wcsrchr
#define sticmp wcsicmp
#define stcmp wcscmp
typedef wstring tstring;
#else
#define stlen strlen
#define stcpy strcpy
#define stcat strcat
#define strchr strrchr
#define sticmp stricmp
#define stcmp strcmp
typedef string tstring;
#endif


HINSTANCE hInstance=0;
HWND hwnd=0;
tstring notepadpath;
HICON hicon=0,hiconsm=0;

class TAutoClose {public: ~TAutoClose() {
  if (hwnd!=0) DestroyWindow(hwnd); hwnd=0;
  if (hicon!=0) DestroyIcon(hicon); hicon=0;
  if (hiconsm!=0) DestroyIcon(hiconsm); hiconsm=0; } } AutoCloser;



int NeedHwnd();

int Error(const TCHAR *fmt, const TCHAR *dat,bool usehwnd=false)
{ const TCHAR *err;
  FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_IGNORE_INSERTS,
    NULL,GetLastError(),MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT),(LPTSTR)&err,0,NULL);
  TCHAR *msg = new TCHAR[stlen(err)+stlen(dat)+stlen(fmt)];
  wsprintf(msg,fmt,err,dat);
  if (usehwnd) NeedHwnd();
  int res = MessageBox(usehwnd?hwnd:NULL,msg,TEXT("Error"),MB_OK|MB_ICONERROR);
  delete[] msg;
  LocalFree((HLOCAL)err);
  return res;
}

int NeedHwnd()
{ if (hwnd!=0) return 1;
  //
  ExtractIconEx(notepadpath.c_str(),0,&hicon,&hiconsm,1);
  WNDCLASSEX wc; ZeroMemory(&wc,sizeof(wc));
  wc.cbSize=sizeof(wc);
  wc.lpfnWndProc=DefWindowProc;
  wc.hInstance=hInstance;
  wc.hIcon=hicon;
  wc.hIconSm=hiconsm;
  wc.lpszClassName=TEXT("ConvertClass");
  ATOM ares = RegisterClassEx(&wc);
  if (ares==0) return Error(TEXT("Internal error registering window class:\r\n%s%s"),TEXT(""),false);
  tstring tit=TEXT("Crlf-notepad");
  hwnd=CreateWindowEx(WS_EX_APPWINDOW,TEXT("ConvertClass"),tit.c_str(),WS_VISIBLE|WS_POPUP,0,0,0,0,NULL,0,hInstance,0);
  if (hwnd==0) return Error(TEXT("Internal error creating window:\r\n%s%s"),TEXT(""),false);
  return 1;
}


tstring ExtractFileName(const tstring fn)
{ const TCHAR *lastslash=strchr(fn.c_str(),'\\'); if (lastslash!=0) lastslash++; else
  { lastslash=strchr(fn.c_str(),'/'); if (lastslash!=0) lastslash++; else
    { lastslash=strchr(fn.c_str(),':'); if (lastslash!=0) lastslash++; else lastslash=fn.c_str();
    }
  }
  return tstring(lastslash);
}

tstring ExtractFilePath(const tstring fn)
{ const TCHAR *lastslash=strchr(fn.c_str(),'\\'); if (lastslash!=0) lastslash++; else
  { lastslash=strchr(fn.c_str(),'/'); if (lastslash!=0) lastslash++; else
    { lastslash=strchr(fn.c_str(),':'); if (lastslash!=0) lastslash++; else lastslash=0;
    }
  }
  if (lastslash==0) return fn;
  else return tstring(fn.c_str(),lastslash-fn.c_str());
}



unsigned int Convert(char *buf, unsigned int size)
{ // PLAN: we will scan forwards counting number of lf, so as to deduce the total
  // file size. Then we scan backwards copying it as necessary.
  int numlf=0;
  for (unsigned int i=0; i<size; i++)
  { if (buf[i]=='\r' && i!=size-1 && buf[i+1]=='\n') i++;
    else if (buf[i]=='\n') numlf++;
  }
  if (numlf==0) return size;
  // we found this many LFs. Therefore, this is how many extra bytes we must add to the file.
  for (unsigned int dst=size+numlf, src=size; src>0; src--,dst--)
  { if (buf[src-1]=='\n' && src>1 && buf[src-2]!='\r') {buf[dst-1]='\n'; buf[dst-2]='\r'; dst--;}
    else buf[dst-1]=buf[src-1];
  }
  if (dst!=0 || src!=0) MessageBox(hwnd,TEXT("There was an internal error converting the file.\r\nI hope nothing was lost! Sorry."),TEXT("Conversion error"),MB_OK);
  return size+numlf;
}


int Install()
{ NeedHwnd();
  tstring msg = TEXT("To install crlf-notepad:\r\n\r\n")
                TEXT("1. Find notepad.exe in your windows or windows\\system32 directory\r\n")
                TEXT("2. Rename it to notepad-orig.exe\r\n")
                TEXT("3. Rename this file to notepad.exe, and place it in that directory.");
  return MessageBox(hwnd,msg.c_str(),TEXT("Crlf-notepad"),MB_OK);
  //int res=MessageBox(hwnd,TEXT("This will install Crlf-notepad"),TEXT("Crlf-notepad Install"),MB_OKCANCEL);
  //if (res==IDCANCEL) return res;
  //
  // make a copy of notepad
  // put ourself in its place
  // create the uninstall batch file
  // add an uninstall key
  //
  //return MessageBox(hwnd,TEXT("Crlf-notepad has been installed.\r\nTo see it in action, load a unixy text file into Notepad!"),TEXT("Crlf-notepad Install"),MB_OKCANCEL);
}

int Uninstall()
{ return 0;
}




int WINAPI WinMain(HINSTANCE hinst,HINSTANCE,LPSTR,int)
{ hInstance=hinst;
  const TCHAR *arg = GetCommandLine(); bool issetup=false;
  if (*arg=='\"')
  { arg++;
    while (*arg!='\"' && *arg!=0)
    { issetup |= ((arg[0]=='s'||arg[0]=='S')&&(arg[1]=='e'||arg[1]=='E')&&(arg[2]=='t'||arg[2]=='T')&&(arg[3]=='u'||arg[3]=='U')&&(arg[4]=='p'||arg[4]=='P'));
      arg++;
    }
    if (*arg=='\"') arg++;
  }
  else
  { while (*arg!=' ')
    { issetup |= ((arg[0]=='s'||arg[0]=='S')&&(arg[1]=='e'||arg[1]=='E')&&(arg[2]=='t'||arg[2]=='T')&&(arg[3]=='u'||arg[3]=='U')&&(arg[4]=='p'||arg[4]=='P'));
      arg++;
    }
  }
  while (*arg==' ') arg++;
  TCHAR mod[MAX_PATH]; GetModuleFileName(hInstance,mod,MAX_PATH);
  tstring path = ExtractFilePath(mod);
  notepadpath = path+TEXT("\\notepad-orig.exe");
  bool notepadexists = (GetFileAttributes(notepadpath.c_str())!=0xFFFFFFFF);
  issetup |= !notepadexists;
  //
  if (issetup) return Install();
  if (sticmp(arg,TEXT("/uninstall"))==0) return Uninstall();
  //
  tstring fn;
  if (*arg=='\"')
  { const TCHAR *start = arg+1, *c=start;
    while (*c!='\"' && *c!=0) c++;
    fn=tstring(start,c-start);
  }
  else
  { const TCHAR *start=arg, *c=arg;
    while (*c!=' ' && *c!=0) c++;
    fn=tstring(start,c-start);
  }
  tstring name=ExtractFileName(fn);

  bool isunix=false;
  HANDLE hf = CreateFile(fn.c_str(),GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,0,NULL);
  if (hf!=INVALID_HANDLE_VALUE)
  { DWORD sizehigh, size=GetFileSize(hf,&sizehigh);
    if (sizehigh==0 && size>2 && size<0x00FFFFFF)
    { if (size>1000) size=1000; unsigned char tst[1000]; tst[size-1]=0;
      DWORD red; ReadFile(hf,tst,size-1,&red,NULL); CloseHandle(hf);
      int numtext=0, numbin=0, numlf=0, numcrlf=0, numhightext=0;
      for (unsigned int i=0; i<size-2; i++)
      { unsigned char c = tst[i];
        if (c>=32 && c<=127) numtext++;
        else if (c>=128 && c<=255) numhightext++;
        else if (c=='\t' || c=='\f') numtext++;
        else if (c=='\r' && tst[i+1]=='\n') {numcrlf++; i++;}
        else if (c=='\n') numlf++;
        else numbin++;
      }
      // heuristics for whether it's a unix file
      isunix = (numlf>1 && numcrlf==0 && numbin==0 && numhightext<numtext/10 && numtext>5);
    }
  }
  
  if (isunix)
  { tstring msg = TEXT("The file ")+name+TEXT(" seems to be a unix text file.\r\nDo you want to convert it to Windows?");
    NeedHwnd();
    int res = MessageBox(hwnd,msg.c_str(),TEXT("Convert text file?"),MB_YESNOCANCEL);
    bool wasreadonly=false;
    if (res==IDYES)
    { DWORD attr = GetFileAttributes(fn.c_str());
      if (attr&FILE_ATTRIBUTE_READONLY)
      { tstring msg = TEXT("The file ")+name+TEXT(" is read-only.\r\nAre you sure you want to convert it?");
        res = MessageBox(hwnd,msg.c_str(),TEXT("Convert text file?"),MB_YESNOCANCEL);
        if (res==IDYES)
        { attr &= ~FILE_ATTRIBUTE_READONLY;
          SetFileAttributes(fn.c_str(),attr);
          wasreadonly=true;
        }
      }
    }
    if (res==IDCANCEL) return 0;
    if (res==IDYES)
    { HANDLE hf = CreateFile(fn.c_str(),GENERIC_READ|GENERIC_WRITE,0,NULL,OPEN_EXISTING,0,NULL);
      if (hf==INVALID_HANDLE_VALUE) Error(TEXT("Unable to open file:\r\n%s\r\nThe file was\r\n%s"),fn.c_str());
      else
      { DWORD size=GetFileSize(hf,NULL), newsize=size;
        HANDLE hmap = CreateFileMapping(hf,NULL,PAGE_READWRITE,0,size*2,NULL);
        if (hmap==NULL) Error(TEXT("Unable to create file mapping:\r\n%s\r\nThe file was\r\n%s"),fn.c_str());
        else
        { char *buf = (char*)MapViewOfFile(hmap,FILE_MAP_WRITE,0,0,size*2);
          if (buf==0) Error(TEXT("Unable to map file:\r\n%s\r\nThe file was\r\n%s"),fn.c_str());
          else
          { newsize = Convert(buf,size);
            UnmapViewOfFile(buf);
          }
          CloseHandle(hmap);
        }
        SetFilePointer(hf,newsize,NULL,FILE_BEGIN); 
        SetEndOfFile(hf); // truncates it!
        CloseHandle(hf);
        if (wasreadonly)
        { DWORD attr = GetFileAttributes(fn.c_str());
          attr|=FILE_ATTRIBUTE_READONLY;
          SetFileAttributes(fn.c_str(),attr);
        }
      }
    }
  }

  tstring newcmd = TEXT("\"")+notepadpath+TEXT("\" ")+arg;
  STARTUPINFO si; ZeroMemory(&si,sizeof(si)); si.cb=sizeof(si);
  PROCESS_INFORMATION pi; ZeroMemory(&pi,sizeof(pi));
  BOOL res = CreateProcess(NULL,(TCHAR*)newcmd.c_str(),NULL,NULL,FALSE,0,NULL,NULL,&si,&pi);
  if (res) {CloseHandle(pi.hProcess); CloseHandle(pi.hThread);}
  else Error(TEXT("Unable to launch notepad:\r\n%s\r\nThe command line was\r\n%s"),newcmd.c_str());
  return 0;
}
